__version__ = "0.1.2"
__title__ = "mia"
__author__ = "Bogdan Kulynych, Mohammad Yaghini"
__email__ = "hello@bogdankulynych.me, mohammad.yaghini@epfl.ch"
__copyright__ = "2018 by Bogdan Kulynych (EPFL SPRING Lab)"
__description__ = "A library for running membership inference attacks against ML models"
__url__ = "https://github.com/spring-epfl/mia"
__license__ = "MIT"
